# -*- coding: utf-8 -*-
# Default Python Libraries
from concurrent.futures import ThreadPoolExecutor
from importlib import import_module
from typing import Dict, Generator, Tuple, Type, TypeVar

# Third Party Libraries
import sseclient

# AristaFlow REST Libraries
from af_org_model_manager.models.client_session_details import ClientSessionDetails

from .rest_helper import RestPackage, RestPackageInstance, RestPackageRegistry


T = TypeVar("T")


class ServiceProvider(object):
    """Per user session service provider caching service stubs and providing authentication data to them."""

    __rest_package_registry: RestPackageRegistry = None
    # package to package instance map
    __rest_packages: [RestPackage, RestPackageInstance] = None
    # cached service stubs
    __services: [type, object] = None
    # authentication
    __csd: ClientSessionDetails = None
    # thread pool for sse handling
    __thread_pool: ThreadPoolExecutor = None

    def __init__(
        self,
        rest_package_registry: RestPackageRegistry,
        thread_pool: ThreadPoolExecutor,
    ):
        self.__rest_package_registry = rest_package_registry
        self.__rest_packages = {}
        self.__services = {}
        self.__thread_pool = thread_pool

    def __get_package_instance(self, service_type: Type) -> RestPackageInstance:
        """Returns the (cached) ApiClient instance for the package of the given service_type.
        Note, that service_type may also be a model class.
        """
        # find the package description and package instance
        pkg = self.__rest_package_registry.get_rest_package(service_type)
        pkg_instance: RestPackageInstance = None
        if pkg in self.__rest_packages:
            pkg_instance = self.__rest_packages[pkg]
        else:
            pkg_instance = RestPackageInstance(pkg, self.__thread_pool)
            self.__rest_packages[pkg] = pkg_instance

        return pkg_instance

    def get_service(self, service_type: Type[T]) -> T:
        """
        Returns a service instance for the given service type, e.g.
        get_service(InstanceControlApi)
        @param service_type The class of the requested service.
        """
        # return a cached instance
        if service_type in self.__services:
            return self.__services[service_type]

        # get the ApiClient object of the package
        api_client = self.__get_package_instance(service_type).api_client

        # authentication data
        if self.__csd:
            self.__use_authentication(api_client)
        # instantiate the service
        service = service_type(api_client)
        # cache it
        self.__services[service_type] = service
        return service

    def __use_authentication(self, api_client: object):
        """Uses the client session details to set the default header"""
        api_client.set_default_header("x-AF-Security-Token", self.__csd.token)

    def authenticated(self, csd: ClientSessionDetails):
        """set the authentication for all ApiClient objects"""
        if self.__csd is not None:
            raise Exception("Already authenticated")

        self.__csd = csd
        for key in self.__rest_packages:
            inst = self.__rest_packages[key]
            self.__use_authentication(inst.api_client)

    def deserialize(self, data, klass):
        """Deserialize data using the given class of the generated OpenAPI models."""
        return self.__get_package_instance(klass).deserialize(data, klass)

    def serialize(self, obj) -> Dict:
        """Serialize REST model object"""
        klass = type(obj)
        return self.__get_package_instance(klass).serialize(obj)

    def get_arista_flow_service_api(self, klass: Type):
        """
        Returns the AristaFlowApi object for the endpoint the given klass belongs to.
        """
        package_name = self.__rest_package_registry.get_package_name(klass)
        package_name = package_name + ".api.arista_flow_service_api"
        afsapi_module = import_module(package_name)
        afsapi_class = afsapi_module.__getattribute__("AristaFlowServiceApi")
        return self.get_service(afsapi_class)

    @property
    def thread_pool(self) -> ThreadPoolExecutor:
        return self.__thread_pool

    def connect_sse(self, klass: Type) -> Tuple[str, sseclient.SSEClient]:
        """
        Creates a new SSE connection for the given endpoint
        """
        pkg = self.__rest_package_registry.get_rest_package(klass)
        service_type = pkg.service_type_name
        host = self.__rest_package_registry.af_conf.get_host(service_type)
        register_sse_endpoint = host + "/sse"
        kwargs = {}
        kwargs["headers"] = {}
        kwargs["headers"]["x-AF-Security-Token"] = self.__csd.token
        kwargs["verify"] = self.__rest_package_registry.af_conf.verify_ssl

        sse: sseclient.SSEClient = sseclient.SSEClient(register_sse_endpoint, **kwargs)
        initial = next(sse)
        return (initial.data, sse)

    @property
    def csd(self):
        """
        Returns the client session details
        """
        return self.__csd



# Default Python Libraries
import logging
from concurrent.futures import ThreadPoolExecutor

from .client_service import AristaFlowClientService
from .configuration import Configuration
from .rest_helper import RestPackageRegistry
from .service_provider import ServiceProvider


class AristaFlowClientPlatform(object):
    """Entry point to the AristaFlow Python Client framework."""

    # thread pool for async requests
    __thread_pool: ThreadPoolExecutor = None

    def __init__(self, configuration: Configuration):
        self.configuration = configuration
        self.__client_services: [str, AristaFlowClientService] = {}
        self.__rest_package_registry = RestPackageRegistry(configuration)
        self.__thread_pool = ThreadPoolExecutor()

    def get_client_service(self, user_session: str = "python_default_session"):
        """
        :return: AristaFlowClientService The client service for the given user session
        """
        if user_session in self.__client_services:
            return self.__client_services[user_session]
        cs = AristaFlowClientService(
            self.configuration,
            user_session,
            ServiceProvider(
                self.__rest_package_registry, self.__thread_pool
            ),
        )
        self.__client_services[user_session] = cs
        return cs

    def disconnect(self):
        try:
            for user_session in self.__client_services:
                self.__client_services[user_session].disconnect()
            # shutdown thread pool
            self.__thread_pool.shutdown(wait=False, cancel_futures=True)
        except BaseException as ex:
            logging.warning('Exception on disconnect', ex)



# -*- coding: utf-8 -*-
# Default Python Libraries
import re
from datetime import datetime


class Version(object):
    """Parsing and comparing version strings"""

    __pattern: re = re.compile("(\\d+)(?:\\.(\\d+))?(?:\\.(\\d+))?(?:\\.(\\d+))?(?:\\.(\\d+))?")

    def key(self, v: str) -> int:
        if not (v):
            return 0

        m: re = self.__pattern.match(v)
        # invalid version string
        if m is None:
            return 0

        gs = m.groups()
        key = 0
        places = len(gs)
        for i in range(places):
            p = gs[i]
            if p is None:
                break
            if len(p) > 3:
                # ugh... unsupported
                return 0
            key += int(p) * 10 ** (places - i)

        # print(f"{v} = {key}")
        return key


VERSION = Version()

# print(v.compare("1.5.0", "1.5.0"))
# vs = ["999.999.999.999.999", "1.5.0", "2.3.555.0", "2.3.8", "5.1", "3", "0.9.6", "2.3.8.1", "2.3.5"]
# print(sorted(vs, key=v.key))


def TO_BPM_DATE(dt: datetime) -> int:
    """Converts a datetime to a BPM platform date/time value"""
    if not (dt):
        return 0
    return int(dt.timestamp() * 1000)


def FROM_BPM_DATE(timestamp: int) -> datetime:
    """Converts a date / time value from the BPM platform to a datetime object"""
    if timestamp == 0:
        return None
    return datetime.fromtimestamp(timestamp / 1e3)


class OrgUtils(object):
    """Org model related utilities"""

    @classmethod
    def summarize_qa_list(cls, *qas):
        """Returns a summarizing string for the given qas"""
        if not (qas):
            return ""
        summary = ""
        for agent in qas:
            name = (
                agent.agent_name
                if agent.agent_name == agent.org_pos_name
                else f"{agent.agent_name} ({agent.org_pos_name})"
            )
            if summary == "":
                summary = name
            else:
                summary = summary + ", " + name

        return summary

    @classmethod
    def build_staff_assignment_rule_for_agent(cls, agent_id: int, org_pos_id: int) -> str:
        """
        Creates a staff assignment rule selecting exactly the given agent.
        """
        return f"Agent(id={agent_id}) AND OrgPosition(id={org_pos_id})"



import asyncio
import json
import threading
import traceback
from asyncio import sleep

import sseclient
from af_remote_html_runtime_manager import SynchronousActivityStartingApi, ActivityExecData, ActivitySseCallbackData
from af_worklist_manager import AfActivityReference, WorklistItem

from .service_provider import ServiceProvider
from .abstract_service import AbstractService

from af_remote_html_runtime_manager.api.runtime_manager_api import RuntimeManagerApi
from af_remote_html_runtime_manager.models.activity_rest_callback_data import (
    ActivityRestCallbackData,
)
from af_remote_html_runtime_manager.models.ebp_instance_reference import EbpInstanceReference
from af_remote_html_runtime_manager.models.gui_context import GuiContext
from .html_gui_context import HtmlGuiContext


class HtmlSignalHandler(object):
    """
    Default signal handler, extend this class for customizations
    """

    _gui_context: HtmlGuiContext = None

    @property
    def gui_context(self) -> HtmlGuiContext:
        return self._gui_context

    @gui_context.setter
    def gui_context(self, gui_context: HtmlGuiContext):
        self._gui_context = gui_context

    def signal(self, activity_signal: str, msg: ActivityExecData):
        """
        To be overridden by sub classes
        The activity has terminated with the given signal which is one of:
        activity-closed
        activity-suspended
        activity-timed-suspended
        activity-resumed
        activity-reset
        activity-signalled
        activity-failed
        :param str activity_signal: The signal
        :param ActivityExecData msg: The execution message
        :return bool: True, for acknowledging the signal (default).
        """


class RemoteHtmlService(AbstractService):
    """
    Helper methods for starting activities via RemoteHTMLRuntimeManager
    """
    __push_sse_client : sseclient.SSEClient = None
    __push_sse_connection_id: str = None
    __value_lock: threading.Lock = None

    def __init__(self, service_provider: ServiceProvider):
        AbstractService.__init__(self, service_provider)
        self.__signal_handlers = {}
        self.__value_lock = threading.Lock()

    @staticmethod
    def is_html_activity(item: WorklistItem) -> bool:
        """
        Returns true, if the given worklist item can be executed as Remote HTML activity
        """
        return (
            item.act_ref.gui_context_id == "HTMLContext"
            or item.act_ref.executable_component_name == "de.aristaflow.form.Form"
            or item.act_ref.executable_component_name == "de.aristaflow.form.GeneratedForm"
        )

    def start_html_activity(self, item: WorklistItem, callback_url: str = None,
                            signal_handler: HtmlSignalHandler = None):
        """
        Starts the given HTML GUI worklist item using the Remote HTML Runtime Manager
        :param WorklistItem item: The worklist item to be started
        :param str callback_url: Use this and the URL will be called on activity signals
        :param HtmlSignalHandler signal_handler: Use this to receive signals via SSE connection
        """
        if item is None:
            raise Exception("No worklist item provided")
        # accept user form and HTMLContext based activities
        if not (self.is_html_activity(item)):
            raise Exception(f"Not an HTML activity: {item.act_ref.gui_context_id}")
        if item.state == "STARTED":
            raise Exception("Item is already started")
        sas = self.get_html_activity_starting()
        gc: GuiContext
        ar: AfActivityReference = item.act_ref
        # print('Starting activity...')
        ebp_ir: EbpInstanceReference = EbpInstanceReference(
            ar.type,
            ar.instance_id,
            ar.instance_log_id,
            ar.base_template_id,
            ar.node_id,
            ar.node_iteration,
            ar.execution_manager_uris,
            ar.runtime_manager_uris,
        )
        # "AVAILABLE", "ASSIGNED", "STARTED", "SUSPENDED", "ENQUIRED"
        if item.state == "AVAILABLE" or item.state == "ASSIGNED":
            if callback_url is not None:
                cb_data = ActivityRestCallbackData(
                    sub_class="ActivityRestCallbackData",
                    notification_callback=callback_url,
                    activity=ebp_ir,
                )
                gc = sas.start_activity_callback(body=cb_data)
            elif signal_handler is not None:
                # ensure SSE connection
                self._register_sse()
                sse_data = ActivitySseCallbackData(
                    sub_class="ActivitySseCallbackData",
                    sse_conn=self.__push_sse_connection_id,
                    activity=ebp_ir,
                )
                gc = sas.start_activity_sse(body=sse_data)
            else:
                gc = sas.start_activity(body=ebp_ir)
        else:
            if callback_url is not None:
                cb_data = ActivityRestCallbackData(
                    sub_class="ActivityRestCallbackData",
                    notification_callback=callback_url,
                    activity=ebp_ir,
                )
                gc = sas.resume_activity_callback(body=cb_data)
            elif signal_handler is not None:
                # ensure SSE connection
                self._register_sse()
                sse_data = ActivitySseCallbackData(
                    sub_class="ActivitySseCallbackData",
                    sse_conn=self.__push_sse_connection_id,
                    activity=ebp_ir,
                )
                gc = sas.resume_activity_sse(body=sse_data)
            else:
                gc = sas.resume_activity(body=ebp_ir)
        hgc = HtmlGuiContext(gc)
        if signal_handler is not None:
            signal_handler.gui_context = hgc
            self.__signal_handlers[gc.session_id] = signal_handler
        return hgc

    def get_html_activity_starting(self) -> SynchronousActivityStartingApi:
        """
        Returns the Remote HTML Runtime Manager Synchronous Activity Starting, ensuring logon to the Runtime Manager
        """
        sas: SynchronousActivityStartingApi = self._service_provider.get_service(SynchronousActivityStartingApi)
        rm: RuntimeManagerApi = self._service_provider.get_service(RuntimeManagerApi)
        # always logon again, since the server might have been restarted in the meantime
        rm.logon(body=self._service_provider.csd)
        return sas

    def _register_sse(self):
        """
        Registers at the runtime service' SSE endpoint and starts the event handling loop.
        """
        if self.__push_sse_client is not None:
            return
        print('RemoteHtmlService register SSE')
        self.__push_sse_connection_id, self.__push_sse_client = self._service_provider.connect_sse(
            SynchronousActivityStartingApi
        )
        print('RemoteHtmlService register SSE, got id ' + self.__push_sse_connection_id)
        self._service_provider.thread_pool.submit(self._process_push_updates)

    def _process_push_updates(self):
        print('RemoteHtmlService starting push update processing')
        """
        Coroutine retrieving SSE push notifications for the activities, handling registration and reconnects
        """
        while not self._disconnected:
            try:
                if self.__push_sse_client is None:
                    print("Establishing SSE connection...")
                    (
                        self.__push_sse_connection_id,
                        self.__push_sse_client,
                    ) = self._service_provider.connect_sse(SynchronousActivityStartingApi)
                print(f"RemoteHtmlService SSE connection established, id is {self.__push_sse_connection_id}")
                while True:
                    for event in self.__push_sse_client:
                        print(f"Event {event.event} received: {event.data}")
                        activity_signal: str = None
                        if event.event == 'SseConnectionEstablished':
                            # print('SSE session was re-established, re-registering..')
                            self.__push_sse_connection_id = event.data
                            # TODO notify waiters / lack of re-registration possibility?
                            print("Runtime SSE connection was re-established!!!")
                        elif event.event == 'activity-closed':
                            activity_signal = 'activity-closed'
                        elif event.event == 'activity-suspended':
                            activity_signal = 'activity-suspended'
                        elif event.event == 'activity-timed-suspended':
                            activity_signal = 'activity-timed-suspended'
                        elif event.event == 'activity-resumed':
                            activity_signal = 'activity-resumed'
                        elif event.event == 'activity-reset':
                            activity_signal = 'activity-reset'
                        elif event.event == 'activity-signalled':
                            activity_signal = 'activity-signalled'
                        elif event.event == 'activity-failed':
                            activity_signal = 'activity-failed'
                        else:
                            print(f"Unknown worklist SSE push event {event.event} received")

                        if activity_signal is not None:
                            print("Remote HTML activity signal received")
                            try:
                                data_dict = json.loads(event.data)
                                data: ActivityExecData = self._service_provider.deserialize(
                                    data_dict, ActivityExecData
                                )
                                # call the listeners
                                if data.session_id in self.__signal_handlers:
                                    try:
                                        signal_handler = self.__signal_handlers[data.session_id]
                                        signal_handler.signal(activity_signal, data)
                                    except Exception as e:
                                        print(
                                            f"Caught an exception signalling activity {data.session_id}, {e}"
                                        )
                                        traceback.print_exc()
                                # self._notify_worklist_update_listeners(data)
                            except Exception as e:
                                print("Couldn't deserialize and apply update: ", event, e)
                    pass
            except ConnectionError:
                # re-establish connection after some wait time
                print("SSE disconnected...")
                sleep(self.__af_conf.sse_connect_retry_wait)
            except Exception as e:
                print("Unknown exception caught during SSE handling", e.__class__)
                raise
            finally:
                self.__push_sse_client = None

    def disconnect(self):
        AbstractService.disconnect(self)
        # close the SSE connection if any
        if self.__push_sse_client:
            self.__push_sse_client.close()



# -*- coding: utf-8 -*-
# Default Python Libraries
import asyncio
import json
import traceback
import warnings
from asyncio import sleep
from threading import Lock
from typing import Generator, List, Set, Union

# Third Party Libraries
from requests import ConnectionError

# AristaFlow REST Libraries
from af_worklist_manager.api.inc_client_worklists_api import IncClientWorklistsApi
from af_worklist_manager.api.inc_worklist_update_api import IncWorklistUpdateApi
from af_worklist_manager.api.worklist_update_manager_api import WorklistUpdateManagerApi
from af_worklist_manager.models.client_worklist_item import ClientWorklistItem
from af_worklist_manager.models.client_worklist_item_update import ClientWorklistItemUpdate
from af_worklist_manager.models.client_worklist_rest_callback_data import (
    ClientWorklistRestCallbackData,
)
from af_worklist_manager.models.client_worklist_sse_callback_data import (
    ClientWorklistSseCallbackData,
)
from af_worklist_manager.models.inc_client_worklist_data import IncClientWorklistData
from af_worklist_manager.models.inc_worklist_update_data import IncWorklistUpdateData
from af_worklist_manager.models.initial_inc_client_worklist_data import InitialIncClientWorklistData
from af_worklist_manager.models.initial_inc_worklist_update_data import InitialIncWorklistUpdateData
from af_worklist_manager.models.update_interval import UpdateInterval
from af_worklist_manager.models.worklist_revision import WorklistRevision
from af_worklist_manager.models.worklist_update import WorklistUpdate
from af_worklist_manager.models.worklist_update_configuration import WorklistUpdateConfiguration

from af_execution_manager import ActivityStartingApi
from af_worklist_manager import WorklistItem, AfActivityReference
from af_runtime_service import EbpInstanceReference

from .abstract_service import AbstractService
from .configuration import Configuration
from .service_provider import ServiceProvider
from .worklist_model import Worklist


# signature for worklist updates
def __update_listener(updates: List[ClientWorklistItemUpdate]):
    pass


def __worklist_callback_provider(worklist: Worklist) -> str:
    return None


WorklistUpdateListener = type(__update_listener)
"""
    Type for callback function for retrieving worklist updates
"""

WorklistCallbackProvider = type(__worklist_callback_provider)
"""
    Type for callback function for providing worklist callback URLs
"""


class WorklistService(AbstractService):
    # The fetch size for incremental worklists / updates. If None, the
    # server-side default will be used
    fetch_count: int = None
    __worklist: Worklist = None
    __items: List[ClientWorklistItem] = None
    __worklist_callback: str = None
    __worklist_callback_provider: WorklistCallbackProvider = None
    __worklist_update_listeners: Set[WorklistUpdateListener] = None
    __af_conf: Configuration = None
    __push_sse_client = None
    __value_lock: Lock = None

    def __init__(self, service_provider: ServiceProvider, configuration: Configuration):
        self.__items = []
        self.__worklist_update_listeners = set()
        self.__af_conf = configuration
        self.__value_lock = Lock()
        super().__init__(service_provider=service_provider)

    def create_worklist_update_configuration(self, push: bool) -> WorklistUpdateConfiguration:
        """Creates a default worklist update configuration
        :param bool push: If set to true, create a worklist update configuration for push notifications.
        """
        update_intervals: list[UpdateInterval] = []
        if push:
            update_intervals.append(UpdateInterval(0, 200))

        # worklistFilter: NO_TL or TL_ONLY
        # notify_only: if set to true, SSE push will send notifications instead of the updates themselves
        wuc = WorklistUpdateConfiguration(
            update_mode_threshold=0,
            update_intervals=update_intervals,
            worklist_filter="NO_TL",
            notify_only=False,
        )
        return wuc

    def get_worklist(self, worklist_callback: str = None, worklist_callback_provider: WorklistCallbackProvider = None,
                     skip_update: bool = False, skip_callback: bool = False) -> List[ClientWorklistItem]:
        """
        Updates and returns the worklist of the current user
        :param str worklist_callback: Optionally an URL implementing the callback endpoint for push updates
        :param WorklistCallbackProvider worklist_callback_provider: Optionally a callback for generating
            callback URLs for push updates
        :param bool skip_update: Skip the update if already connected but no callback options are enabled.
        """
        # use a lock to prevent concurrent initialization of the worklist
        with self.__value_lock:
            if self.__worklist is None:

                wum: WorklistUpdateManagerApi = self._service_provider.get_service(
                    WorklistUpdateManagerApi
                )
                update_conf: WorklistUpdateConfiguration = self.create_worklist_update_configuration(True)
                wlit: InitialIncClientWorklistData
                if self.fetch_count is not None:
                    wlit = wum.logon_and_create_client_worklist(body=update_conf, count=self.fetch_count)
                else:
                    wlit = wum.logon_and_create_client_worklist(body=update_conf)

                # currently no items in the worklist
                if wlit is None:
                    return self.__items

                self.__iterate(self.__items, wlit)

                # remember the current worklist meta data
                self.__worklist = Worklist(
                    wlit.worklist_id, wlit.revision, wlit.client_worklist_id, update_conf, wlit.agent
                )

                if not skip_callback:
                    self.enable_callback_updates(worklist_callback)

                return self.__items
        # /value_lock end

        # self.__worklist is not none here
        # simply return the items if push notifications are enabled
        if worklist_callback is not None or worklist_callback_provider is not None:
            # ensure the callback is registered
            self.enable_callback_updates(worklist_callback=worklist_callback,
                                         worklist_callback_provider=worklist_callback_provider)
            return self.__items
        # simply return the items if push notifications are enabled
        if self.__worklist_callback is not None or self.__worklist_callback_provider is not None:
            # ensure the callback is still registered
            self.enable_callback_updates(worklist_callback=self.__worklist_callback,
                                         worklist_callback_provider=self.__worklist_callback_provider)
            return self.__items
        if self.__push_sse_client:
            return self.__items
        # perform update
        if not skip_update:
            return self.update_worklist()
        else:
            return self.__items

    def force_worklist_reload(self) -> List[ClientWorklistItem]:
        """
        Forces the reworklist to be reloaded and registrations for push updates, if any, renewed.
        """
        if self.__worklist is None:
            return self.get_worklist()

        wum: WorklistUpdateManagerApi = self._service_provider.get_service(
            WorklistUpdateManagerApi
        )
        worklist_callback = self.__worklist_callback
        if not worklist_callback and self.__worklist_callback_provider:
            worklist_callback = self.__worklist_callback_provider()

        # if callback are enabled, de-register and re-register
        if worklist_callback:
            is_registered = wum.is_registered_for_updates(self.__worklist.worklist_id,
                                                          self.__worklist.client_worklist_id)
            if is_registered:
                wum.unregister_client_worklist_callback(worklist_id=self.__worklist.worklist_id,
                                                        client_worklist_id=self.__worklist.client_worklist_id)
            self._clear_worklist()
            # register again
            self.enable_callback_updates(worklist_callback=self.__worklist_callback,
                                         worklist_callback_provider=self.__worklist_callback_provider,
                                         client_worklist_id=self.__worklist.client_worklist_id)
        # if SSE is enabled, close the connection and re-enable
        elif self.__push_sse_client:
            self.__push_sse_client.close()
            self.__push_sse_client = None
            self._clear_worklist()
            self.pull_and_apply_worklist_updates()
            self.enable_push_updates()
        # no push mechanism: simply clear the worklist and re-fetch it
        else:
            self._clear_worklist()
            self.update_worklist()

    def _clear_worklist(self):
        """
        Clears the worklist, i.e. empties the items list and resets the revision information.
        """
        self.__items.clear()
        self.__worklist.revision.update_count = 0
        self.__worklist.revision.initialisation_date = 1

    def enable_callback_updates(self, worklist_callback: str = None,
                                worklist_callback_provider: WorklistCallbackProvider = None,
                                client_worklist_id: int = None):
        """
        Enable automatic worklist updates using SSE push notifications.
        """
        if worklist_callback is not None or worklist_callback_provider is not None:
            wum: WorklistUpdateManagerApi = self._service_provider.get_service(
                WorklistUpdateManagerApi
            )
            is_registered = wum.is_registered_for_updates(self.__worklist.worklist_id, self.__worklist.client_worklist_id)
            # print(f'Callback for {self.__worklist.worklist_id} registration status: {is_registered}')
            if not is_registered:
                if client_worklist_id is None:
                    client_worklist_id = self.__worklist.client_worklist_id
                callback_data = ClientWorklistRestCallbackData(
                    worklist_callback=worklist_callback if worklist_callback else worklist_callback_provider(self.__worklist),
                    sub_class="ClientWorklistRestCallbackData",
                    id=self.__worklist.worklist_id,
                    client_worklist_id=client_worklist_id,
                    agent=self.__worklist.agent,
                    revision=self.__worklist.revision,
                    wu_conf=self.__worklist.wu_conf,
                )
                wum.register_client_worklist_callback(callback_data)
                self.__worklist_callback = worklist_callback
                self.__worklist_callback_provider = worklist_callback_provider

    def enable_push_updates(self):
        """
        Enable automatic worklist updates using SSE push notifications.
        """
        if self.__push_sse_client is not None:
            return
        # ensure the worklist has been fetched once
        # NOTE: this does not help, if the worklist is empty (BPM-3581)
        if self.__worklist is None:
            self.get_worklist()
        if self.__worklist is None:
            warnings.warn(
                "The worklist could not be initialized, probably due to BPM-3581. SSE push will not work."
            )
        self._service_provider.thread_pool.submit(self._process_push_updates)

    def _process_push_updates(self):
        """
        Coroutine retrieving SSE push notifications for the worklist, handling registration and reconnects
        """
        while not self._disconnected:
            print("WorklistService establishing SSE connection...")
            try:
                sse_connection_id, sse_client = self._service_provider.connect_sse(
                    WorklistUpdateManagerApi
                )
                while True:
                    print("WorklistService SSE connection established, registering for worklist push...")
                    callback_data = ClientWorklistSseCallbackData(
                        sse_conn=sse_connection_id,
                        sub_class="ClientWorklistSseCallbackData",
                        id=self.__worklist.worklist_id,
                        client_worklist_id=self.__worklist.client_worklist_id,
                        agent=self.__worklist.agent,
                        revision=self.__worklist.revision,
                        wu_conf=self.__worklist.wu_conf,
                    )
                    wum: WorklistUpdateManagerApi = self._service_provider.get_service(
                        WorklistUpdateManagerApi
                    )
                    wum.register_client_worklist_sse(callback_data)
                    print("Worklist registered for SSE push")
                    self.__push_sse_client = sse_client
                    for event in sse_client:
                        if event.event == "SseConnectionEstablished":
                            # print('SSE session was re-established, re-registering..')
                            callback_data.sse_conn = event.data
                            callback_data.revision = self.__worklist.revision
                            wum.register_client_worklist_sse(callback_data)
                            # print("Worklist registered again for SSE push")
                        elif event.event == "client-worklist-update":
                            # print("Worklist update received")
                            try:
                                update_dict = json.loads(event.data)
                                update: WorklistUpdate = self._service_provider.deserialize(
                                    update_dict, WorklistUpdate
                                )
                                self.apply_worklist_update(update)
                            except Exception as e:
                                print("Couldn't deserialize and apply update: ", event, e)
                        else:
                            print(f"Unknown worklist SSE push event {event.event} received")
            except ConnectionError:
                # re-establish connection after some wait time
                # print("SSE disconnected...")
                sleep(self.__af_conf.sse_connect_retry_wait)
            except Exception as e:
                print("Unknown exception caught during SSE handling", e.__class__)
                traceback.print_exc()
                raise
            finally:
                self.__push_sse_client = None

    def add_update_listener(self, listener: WorklistUpdateListener):
        """
        Adds a listener which is called after a worklist update was received and applied.
        """
        self.__worklist_update_listeners.add(listener)

    def remove_update_listener(self, listener: WorklistUpdateListener):
        """
        Removes the given worklist update listener
        """
        self.__worklist_update_listeners.remove(listener)

    def _notify_worklist_update_listeners(self, updates: List[ClientWorklistItemUpdate]):
        """
        Notifies all registered worklist update listeners
        """
        for listener in self.__worklist_update_listeners:
            try:
                listener(updates)
            except Exception as e:
                print("Caught exception while notifying listener:", e)
                traceback.print_exc()

    def worklist_meta_data(self) -> Worklist:
        """
        Returns the worklist meta data, like ID, current revision etc., don't modify.
        """
        if self.__worklist is None:
            self.get_worklist()
        return self.__worklist

    def __iterate(
        self,
        items: List[ClientWorklistItem],
        inc: Union[InitialIncClientWorklistData, IncClientWorklistData],
    ):
        """Consumes an incremental client worklist until its iterator is used up
        @param items The items list to fill with the update(s)
        @param inc The first or next iteration to consume and append to items.
        """
        if inc is None:
            return
        # append the items
        if inc.items_flat:
            items += inc.items_flat
        else:
            return
        # iterator is used up
        if inc.dropped:
            return

        # fetch next
        inc_cl: IncClientWorklistsApi = self._service_provider.get_service(IncClientWorklistsApi)
        next_it: IncClientWorklistData = inc_cl.inc_client_wl_get_next(inc.inc_wl_id)
        self.__iterate(items, next_it)

    def update_worklist(self) -> List[ClientWorklistItem]:
        """Updates the user's worklist and returns the items"""
        if self.__worklist is None:
            return self.get_worklist()

        if self.__push_sse_client is not None \
                or self.__worklist_callback is not None\
                or self.__worklist_callback_provider is not None:
            return self.__items

        self.pull_and_apply_worklist_updates()

        return self.__items

    def pull_and_apply_worklist_updates(self):
        """
        Pulls worklist updates from the server and applies them to the worklist. Using this method while having
        push updates enabled via SSE or callback, may lead to an inconsistent state in the local list of items.
        """
        if self.__worklist is None:
            return self.get_worklist()
        wu: WorklistUpdateManagerApi = self._service_provider.get_service(WorklistUpdateManagerApi)
        inc_updts: InitialIncWorklistUpdateData = wu.get_worklist_updates(
            self.__worklist.worklist_id,
            body=self.__worklist.revision,
            filter=self.__worklist.wu_conf.worklist_filter,
        )

        if inc_updts is not None:
            updates: List[ClientWorklistItemUpdate] = []
            self.__iterate_updates(updates, inc_updts)
            self.__apply_worklist_updates(
                inc_updts.source_revision, inc_updts.target_revision, updates
            )
            self._notify_worklist_update_listeners(updates)

    # TODO define fitting return value
    def select_activity(
        self, item: WorklistItem #, signal_handler: SignalHandler = None
    ) -> bool:
        """
        Assigns the current user to the provided activity
        """
        asa: ActivityStartingApi = self._service_provider.get_service(
            ActivityStartingApi
        )
        
        ar: AfActivityReference = item.act_ref
        ebp_ir = EbpInstanceReference(
            ar.type,
            ar.instance_id,
            ar.instance_log_id,
            ar.base_template_id,
            ar.node_id,
            ar.node_iteration,
            ar.execution_manager_uris,
            ar.runtime_manager_uris,
        )

        asa.select_activity(body=ebp_ir)

        return True

    def __iterate_updates(
        self, updates: List[ClientWorklistItemUpdate], inc: IncWorklistUpdateData
    ):
        """Consumes the given worklist update iterator and appends all retrieved updates to the provided updates list."""
        if inc is None:
            return
        if inc.item_updates:
            updates += inc.item_updates
        else:
            return
        if inc.dropped:
            return

        # fetch next
        iwua: IncWorklistUpdateApi = self._service_provider.get_service(IncWorklistUpdateApi)
        next_it: IncWorklistUpdateData = iwua.inc_wl_updt_get_next(inc.inc_upd_id)
        self.__iterate_updates(updates, next_it)

    def apply_worklist_update(self, update: WorklistUpdate):
        """ Apply the provided worklist update to the worklist
        """
        self.__apply_worklist_updates(
            update.source_revision,
            update.target_revision,
            update.item_updates,
        )
        # call the listeners
        self._notify_worklist_update_listeners(update.item_updates)

    def __apply_worklist_updates(
        self,
        source_revision: WorklistRevision,
        target_revision: int,
        updates: List[ClientWorklistItemUpdate],
    ):
        """Applies the provided worklist updates to self.__items. Checks the consistency to the source revision,
        and performs a full update if the state does not fit. Sets the new target revision in self.__worklist.
        """
        if (
            self.__worklist.revision.update_count != source_revision.update_count
            or self.__worklist.revision.initialisation_date != source_revision.initialisation_date
        ):
            # out of order update, clear and re-fetch everything
            self.__items.clear()
            self.__worklist = None
            self.get_worklist()
            return

        # print(f'Applying {len(updates)} updates')
        # print(updates)

        for update in updates:
            self.__apply_worklist_update(update)

        # remember the update count for the next update
        self.__worklist.revision.update_count = target_revision

    def __apply_worklist_update(self, update: ClientWorklistItemUpdate):
        """Applies the given update to __items"""
        update_type = update.update_type
        item = update.item
        if update_type == "CHANGED":
            self.__replace_or_add_item(item)
        elif update_type == "ADDED":
            self.__items += [item]
        elif update_type == "REMOVED":
            self.__remove_item(item)
        elif update_type == "ADDED_OR_CHANGED":
            self.__replace_or_add_item(item)
        elif update_type == "REMOVED_OR_NOTHING":
            self.__remove_item(item)
        else:
            raise RuntimeError("Unknown update type: " + update_type)

    def __replace_or_add_item(self, item: ClientWorklistItem):
        """Replaces or adds the given worklist item in self.__items"""
        # print('__replace_or_add_item: __items=', self.__items)
        for i in range(len(self.__items)):
            val = self.__items[i]
            if item.id == val.id:
                self.__items[i] = item
                return
        # not found above, append it
        self.__items.append(item)

    def __remove_item(self, item: ClientWorklistItem):
        """Removes the given worklist item from self.__items"""
        for val in self.__items:
            if item.id == val.id:
                self.__items.remove(val)
                return

    def find_item_by_id(self, item_id: str) -> ClientWorklistItem:
        """Finds a worklist item by its worklist item id. Returns none, if not in the worklist of the user."""
        # print(f'Finding item with id {item_id}')
        self.get_worklist()
        # print(self.__items)
        for item in self.__items:
            if item.id == item_id:
                # print('Found')
                return item
        return None

    def find_item_by_ref(self, ebp_ir: EbpInstanceReference) -> ClientWorklistItem:
        """
        Finds a worklist item by its activity's EBP Instance Reference.
        Returns none, if not in the worklist of the user.
        """
        self.get_worklist()
        for item in self.__items:
            ar: AfActivityReference = item.act_ref
            if ar.instance_id == ebp_ir.instance_id\
                    and ar.instance_log_id == ebp_ir.instance_log_id\
                    and ar.node_id == ebp_ir.node_id\
                    and ar.node_iteration == ebp_ir.node_iteration:
                return item
        return None


    def update_worklist_item(self, item: ClientWorklistItem):
        wum: WorklistUpdateManagerApi = self._service_provider.get_service(
            WorklistUpdateManagerApi
        )
        wum.update_individual_settings(item)

    def disconnect(self):
        AbstractService.disconnect(self)
        # close the SSE connection if any
        if self.__push_sse_client:
            self.__push_sse_client.close()



# Default Python Libraries
from typing import Generator

# AristaFlow REST Libraries
from af_process_manager.api.exec_hist_entry_remote_iterator_rest_api import (
    ExecHistEntryRemoteIteratorRestApi,
)
from af_process_manager.api.execution_history_api import ExecutionHistoryApi
from af_process_manager.models.exec_hist_entry_initial_remote_iterator_data import (
    ExecHistEntryInitialRemoteIteratorData,
)
from af_process_manager.models.execution_history_entry import ExecutionHistoryEntry
from aristaflow.abstract_service import AbstractService


class ExecutionHistoryService(AbstractService):
    """
    Accessing the execution history
    """

    def read_instance_history(self, inst_log_id) -> Generator[ExecutionHistoryEntry, None, None]:
        """
        Returns a generator allowing to read an arbitrary number of execution history entries.
        Close the generator for dropping the remote iterator.
        """
        pm: ExecutionHistoryApi = self._service_provider.get_service(ExecutionHistoryApi)
        eh_iter_api: ExecHistEntryRemoteIteratorRestApi = self._service_provider.get_service(
            ExecHistEntryRemoteIteratorRestApi
        )
        next_iter: ExecHistEntryInitialRemoteIteratorData = pm.read_instance_history(
            inst_log_id, body=[]
        )
        # seems unlikely but still occurred
        if next_iter is None:
            return
        iterator_id = next_iter.iterator_id
        try:
            while next_iter and next_iter.entries:
                for entry in next_iter.entries:
                    yield entry
                if next_iter.dropped:
                    break
                next_iter = eh_iter_api.exec_hist_entry_get_next(iterator_id)
        except GeneratorExit:
            # generator closed
            if iterator_id:
                eh_iter_api.exec_hist_entry_drop(iterator_id)
